#!/usr/bin/env python
#
# jabber interface.. to whatever you like
#
# remark:
# jabber connection code is based on the console sample jabber client from
# python-jabber
#
#
# version 0.3 - 20040407
# BeF <bef@erlangen.ccc.de>
#

## general imports
from string import split,strip,join
import sys,os

import socket			## error handling
import re			## command parsing
import imp			## module loading
import ConfigParser		## config gile parsing
import time			## logging
import traceback		## module import + function execution

#sys.path.insert(1, os.path.join(sys.path[0], '..'))
import jabber

########## CONFIG ##########

# Change this to 0 if you dont want a color xterm
USE_COLOR = 1

plugindir = os.path.dirname(sys.argv[0]) + '/plugins'

######################################
## don't change anything below unless you know what you're doing

## config file parsing
config = ConfigParser.ConfigParser()
config.read(os.path.dirname(sys.argv[0]) + '/jabberif.conf')

Server = config.get('account', 'server')
Username = config.get('account', 'username')
Password = config.get('account','password')
Resource = config.get('account', 'resource')

logfile = config.get('logging', 'logfile')


## plugins

class pluginhandler:
	def __init__(self):
		self.__plugins = {}
		self.__commands = {}
	
	def plugins(self):
		return self.__plugins
	def commands(self):
		return self.__commands

	def load(self, name):
		"""load module"""
		if self.__plugins.has_key(name):
			lh.log("error loading: plugin %s already exists" % name, 'error')
			return 1
			
		if not os.path.isfile(plugindir+'/'+name+'.py'):
			lh.log("error loading: plugin %s doesn't exist" % name, 'error')
			return 1

		fp, pathname, description = imp.find_module(name, [plugindir])
		try:
			module = imp.load_module(name, fp, pathname, description)
			self.__plugins[name]={'name': name, 'module': module,
				'obj': eval('module.'+name+'(self)')}
		except ImportError:
			lh.log("error importing module %s" % name, 'error')
		except Exception, e:
			lh.log("error loading module %s: %s" % (name, str(e)), 'error')
			lh.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')
			if fp:
				fp.close()
			return 1
		if fp:
			fp.close()
		return 0
	
	def register_command(self, module, name, function, help='', args = {}):
		"""register command
			this is usually called by modules to register their new commands
			args can be:
				op_only: 0/1"""
		if self.__commands.has_key(name):
			lh.log("error registering cmd: %s already exists in module %s" % (name, self.__commands[name]['module'].name), 'error')
		else:
			lh.log("registering command /%s from module %s" % (name, module.name), 'module')
			self.__commands[name] = {'name': name, 'module': module, 'function': function,
				'help': help, 'args': args}
				
	def unload(self, name):
		"""unload module"""
		if not self.__plugins.has_key(name):
			lh.log("error unloading: plugin %s is not loaded" % name, 'error')
			return

		module = self.__plugins[name]['obj']

		for cmd in self.__commands.keys():
			if self.__commands[cmd]['module'] is module:
				self.__commands.pop(cmd)
				
		self.__plugins.pop(name)
	
	def send(self, con, to, message):
		"""send message"""
		msg = jabber.Message(to, message)
		msg.setType('chat')
		con.send(msg)
		lh.log(str(to) + ': ' + message, 'outmessage')
		
	def send_back(self, con, msg, message):
		"""send message back to where it came from"""
		self.send(con, msg.getFrom(), message)
	
	def get_config(self, key):
		"""get config file entry from section 'plugins'"""
		return config.get('plugins', key)

	def get_userhandler(self):
		"""return userhandler"""
		return uh
		
	def get_loghandler(self):
		"""return loghandler"""
		return lh
	
	def commandhandler(self, args):
		"""here the commands are relayed to modules"""
		con = args['con']
		msg = args['msg']
		cmd = args['cmd']

		if len(cmd) <= 0:
			return

		if self.__commands.has_key(cmd[0]):
			command = self.__commands[cmd[0]]
			if command['args'].has_key('op_only') and command['args']['op_only'] == 1 and not uh.check_op(msg.getFrom().getStripped()):
				self.send_back(con, msg, 'sorry. you require op status to execute this command')
			else:
				try:
					command['function'](args)
				except Exception, e:
					lh.log("error executing command %s: %s" % (cmd[0], str(e)), 'error')
					lh.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')
					self.send_back(con, msg, 'error executing function')
				
		else:
			self.send_back(con, msg, 'error. try /help')


## users

class userhandler:
	def __init__(self, ops=[]):
		self.userinfo = {}
		for op in ops:
			self.userinfo[op] = {'op': 1}
	
	def set_presence(self, id, jid, type):
		"""set presence of user with id and jid to type"""
		if not self.userinfo.has_key(id):
			self.userinfo[id] = {'presence': {}}
		elif not self.userinfo[id].has_key('presence'):
			self.userinfo[id]['presence'] = {}
		self.userinfo[id]['presence'][jid] = type
	
	def check_op(self, id):
		"""checks for op status"""
		if self.userinfo.has_key(id):
			if self.userinfo[id].has_key('op'):
				if self.userinfo[id]['op'] == 1:
					return 1
		return 0
	
	def get_presence(self, id):
		"""checks for presence"""
		if self.userinfo.has_key(id):
			if self.userinfo[id].has_key('presence'):
				return self.userinfo[id]['presence']
		return {}
	
	def get_all_users(self):
		"""return all cached users available"""
		return self.userinfo.keys()

## log handler

class loghandler:
	def __init__(self):
		self.__logfile = logfile
		self.__log = open(self.__logfile, 'a', 0)
		self.__file_log('*** server start ***')
		
		self.__alltypes = ['error', 'traceback', 'module', 'info', 'connection', 'inmessage', 'outmessage']
		self.__tofile = eval(config.get('logging', 'logtofile'))
		self.__toscreen = eval(config.get('logging', 'logtoscreen'))
		self.__colormap = eval(config.get('logging', 'colormap'))

	def __file_log(self, message):
		self.__log.write(time.strftime("[%x %X]") + ' ' + message + '\n')
	
	def __screen_log(self, message, type):
		color = ''
		if self.__colormap.has_key(type):
			color = self.__colormap[type]
		print self.colorize(message, color)
	
	def log(self, message, type='info'):
		if not type in self.__alltypes:
			self.__screen_log("unknown message type %s" % type, 'error')
			type = 'info'
		if type in self.__tofile:
			self.__file_log(message)
		if type in self.__toscreen:
			self.__screen_log(message, type)
		
	## text colorizer
	def colorize(self, txt, col=''):
		"""Return colorized text"""
		if not USE_COLOR or col == '': return txt
		cols = { 'red':1, 'green':2, 'yellow':3, 'blue':4}
		initcode = '\033[;3'
		endcode  = '\033[0m'
		if type(col) == type(1): 
			return initcode + str(col) + 'm' + txt + endcode
		try: return initcode + str(cols[col]) + 'm' + txt + endcode
		except: return txt
		
	## TODO: actually close the logfile .. at some point
	def close(self):
		self.__log.close()


##################################################


## functions


## message handlers
def messageCB(con, msg):
	"""Called when a message is recieved"""
	if msg.getBody():
		msgbody = strip(msg.getBody())
		lh.log(str(msg.getFrom()) + ' ' + join(re.split('\r?\n', msgbody), '\n\t'), 'inmessage')
		if len(msgbody) > 0:
			if  msgbody[0] == '/':
				ph.commandhandler({'con': con, 'msg': msg, 'cmd': re.split('[\s\t\r\n]+', msgbody[1:])})
			else:
				ph.commandhandler({'con': con, 'msg': msg, 'cmd': ['msg', msgbody]})

def presenceCB(con, prs):
	"""Called when a presence is recieved"""
	who = prs.getFrom()
	type = prs.getType()
	if type == None: type = 'available'

	elif type == 'unsubscribe':
		con.send(jabber.Presence(to=str(who), type='unsubscribed'))
		con.send(jabber.Presence(to=str(who), type='unsubscribe'))
	
	uh.set_presence(who.getStripped(), str(who), type)


def iqCB(con,iq):
	"""Called when an iq is recieved, we just let the library handle it at the moment"""
	pass

def disconnectedCB(con):
	lh.log("Ouch, network error", 'error')
	sys.exit(1)




##################################################

## logging
lh = loghandler()

## load user info
uh = userhandler(ops=re.split('[,\s\n\r]+', config.get('users','ops')))

## load plugins
ph = pluginhandler()

for plugin in eval(config.get('plugins','startup')):
	ph.load(plugin)

##################################################


## non-ssl connect
#con = jabber.Client(host=Server,debug=1 ,log=sys.stderr)
## ssl connect
con = jabber.Client(host=Server, port=5223, connection=jabber.xmlstream.TCP_SSL)

try:
	con.connect()
except IOError, e:
	lh.log("Couldn't connect: %s" % e, 'error')
	sys.exit(0)
else:
	lh.log("Connected",'connection')

## handlers
con.registerHandler('message', messageCB)
con.registerHandler('presence', presenceCB)
con.registerHandler('iq', iqCB)
con.setDisconnectHandler(disconnectedCB)


## login
lh.log("Attempting to log in...", 'connection')

if con.auth(Username,Password,Resource):
	lh.log("Logged in as %s to server %s" % ( Username, Server), 'connection')
else:
	lg.log("eek -> %s %s" % (con.lastErr, con.lastErrCode), 'error')
	sys.exit(1)

## roster/presence init
lh.log("Requesting Roster Info" , 'connection')
con.requestRoster()
con.sendInitPresence()

JID = Username + '@' + Server + '/' + Resource

## loop forever
while(1):
	try:
		con.process(1)
	except socket.sslerror:
		lh.log("end of fnord.", 'info')
		sys.exit(0)
#	except:
#		print "unknown error"
#		sys.exit(1)
    
        
